"""Define a client to interact with the Ambient Weather APIs."""
from typing import Optional

from aiohttp import ClientSession

from .api import API
from .websocket import Websocket

DEFAULT_API_VERSION = 1


class Client:  # pylint: disable=too-few-public-methods
    """Define the client."""

    def __init__(
        self,
        api_key: str,
        application_key: str,
        *,
        api_version: int = DEFAULT_API_VERSION,
        session: Optional[ClientSession] = None,
    ) -> None:
        """Initialize."""
        self.api = API(application_key, api_key, api_version, session)
        self.websocket = Websocket(application_key, api_key, api_version)



"""Define an object to interact with the REST API."""
import asyncio
from datetime import date
from typing import Any, Dict, List, Optional

from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientError

from .errors import RequestError

REST_API_BASE: str = "https://api.ambientweather.net"

DEFAULT_LIMIT: int = 288
DEFAULT_TIMEOUT: int = 10


class API:
    """Define the API object."""

    def __init__(
        self,
        application_key: str,
        api_key: str,
        api_version: int,
        session: Optional[ClientSession] = None,
    ) -> None:
        """Initialize."""
        self._api_key: str = api_key
        self._api_version: int = api_version
        self._application_key: str = application_key
        self._session: Optional[ClientSession] = session

    async def _request(
        self, method: str, endpoint: str, **kwargs: Dict[str, Any]
    ) -> List[Dict[str, Any]]:
        """Make a request against the API."""
        # In order to deal with Ambient's fairly aggressive rate limiting, we
        # pause for a second before continuing in case any requests came before
        # this.
        # https://ambientweather.docs.apiary.io/#introduction/rate-limiting
        await asyncio.sleep(1)

        url = f"{REST_API_BASE}/v{self._api_version}/{endpoint}"

        kwargs.setdefault("params", {})
        kwargs["params"]["apiKey"] = self._api_key
        kwargs["params"]["applicationKey"] = self._application_key

        use_running_session = self._session and not self._session.closed

        if use_running_session:
            session = self._session
        else:
            session = ClientSession(timeout=ClientTimeout(total=DEFAULT_TIMEOUT))

        assert session

        data: List[Dict[str, Any]] = []

        try:
            async with session.request(method, url, **kwargs) as resp:
                resp.raise_for_status()
                data = await resp.json()
        except ClientError as err:
            raise RequestError(f"Error requesting data from {url}: {err}") from err
        finally:
            if not use_running_session:
                await session.close()

        return data

    async def get_devices(self) -> List[Dict[str, Any]]:
        """Get all devices associated with an API key."""
        return await self._request("get", "devices")

    async def get_device_details(
        self,
        mac_address: str,
        *,
        end_date: Optional[date] = None,
        limit: int = DEFAULT_LIMIT,
    ) -> list:
        """Get details of a device by MAC address."""
        params: Dict[str, Any] = {"limit": limit}
        if end_date:
            params["endDate"] = end_date.isoformat()

        return await self._request("get", f"devices/{mac_address}", params=params)



"""Define an object to interact with the Websocket API."""
import asyncio
from typing import Any, Awaitable, Callable, Dict, Optional

from aiohttp.client_exceptions import ClientConnectionError, ClientOSError
from socketio import AsyncClient
from socketio.exceptions import ConnectionError as SIOConnectionError, SocketIOError

from .const import LOGGER
from .errors import WebsocketError

DEFAULT_WATCHDOG_TIMEOUT = 900

WEBSOCKET_API_BASE: str = "https://api.ambientweather.net"


class WebsocketWatchdog:
    """Define a watchdog to kick the websocket connection at intervals."""

    def __init__(
        self,
        action: Callable[..., Awaitable],
        *,
        timeout_seconds: int = DEFAULT_WATCHDOG_TIMEOUT,
    ):
        """Initialize."""
        self._action: Callable[..., Awaitable] = action
        self._loop = asyncio.get_event_loop()
        self._timer_task: Optional[asyncio.TimerHandle] = None
        self._timeout = timeout_seconds

    def cancel(self) -> None:
        """Cancel the watchdog."""
        if self._timer_task:
            self._timer_task.cancel()
            self._timer_task = None

    async def on_expire(self) -> None:
        """Log and act when the watchdog expires."""
        LOGGER.info("Watchdog expired – calling %s", self._action.__name__)
        await self._action()

    async def trigger(self) -> None:
        """Trigger the watchdog."""
        LOGGER.info("Watchdog triggered – sleeping for %s seconds", self._timeout)

        if self._timer_task:
            self._timer_task.cancel()

        self._timer_task = self._loop.call_later(
            self._timeout, lambda: asyncio.create_task(self.on_expire())
        )


class Websocket:
    """Define the websocket."""

    def __init__(self, application_key: str, api_key: str, api_version: int) -> None:
        """Initialize."""
        self._api_key: str = api_key
        self._api_version: int = api_version
        self._app_key: str = application_key
        self._async_user_connect_handler: Optional[Callable[..., Awaitable]] = None
        self._sio: AsyncClient = AsyncClient(logger=LOGGER, engineio_logger=LOGGER)
        self._user_connect_handler: Optional[Callable] = None
        self._watchdog: WebsocketWatchdog = WebsocketWatchdog(self.reconnect)

    async def _init_connection(self) -> None:
        """Perform automatic initialization upon connecting."""
        await self._sio.emit("subscribe", {"apiKeys": [self._api_key]})
        await self._watchdog.trigger()

        if self._async_user_connect_handler:
            await self._async_user_connect_handler()
        elif self._user_connect_handler:
            self._user_connect_handler()

    def async_on_connect(self, target: Callable[..., Awaitable]) -> None:
        """Define a coroutine to be called when connecting."""
        self._async_user_connect_handler = target
        self._user_connect_handler = None

    def on_connect(self, target: Callable) -> None:
        """Define a method to be called when connecting."""
        self._async_user_connect_handler = None
        self._user_connect_handler = target

    def async_on_data(self, target: Callable[..., Awaitable]) -> None:  # noqa: D202
        """Define a coroutine to be called when data is received."""

        async def _async_on_data(data: Dict[str, Any]) -> None:
            """Act on the data."""
            await self._watchdog.trigger()
            await target(data)

        self._sio.on("data", _async_on_data)

    def on_data(self, target: Callable) -> None:  # noqa: D202
        """Define a method to be called when data is received."""

        async def _async_on_data(data: Dict[str, Any]) -> None:
            """Act on the data."""
            await self._watchdog.trigger()
            target(data)

        self._sio.on("data", _async_on_data)

    def async_on_disconnect(self, target: Callable[..., Awaitable]) -> None:
        """Define a coroutine to be called when disconnecting."""
        self._sio.on("disconnect", target)

    def on_disconnect(self, target: Callable) -> None:
        """Define a method to be called when disconnecting."""
        self._sio.on("disconnect", target)

    def async_on_subscribed(
        self, target: Callable[..., Awaitable]
    ) -> None:  # noqa: D202
        """Define a coroutine to be called when subscribed."""

        async def _async_on_subscribed(data: Dict[str, Any]) -> None:
            """Act on subscribe."""
            await self._watchdog.trigger()
            await target(data)

        self._sio.on("subscribed", _async_on_subscribed)

    def on_subscribed(self, target: Callable) -> None:  # noqa: D202
        """Define a method to be called when subscribed."""

        async def _async_on_subscribed(data: Dict[str, Any]) -> None:
            """Act on subscribe."""
            await self._watchdog.trigger()
            target(data)

        self._sio.on("subscribed", _async_on_subscribed)

    async def connect(self) -> None:
        """Connect to the socket."""
        try:
            self._sio.on("connect", self._init_connection)
            await self._sio.connect(
                f"{WEBSOCKET_API_BASE}/?api={self._api_version}&applicationKey={self._app_key}",
                transports=["websocket"],
            )
        except (
            ClientConnectionError,
            ClientOSError,
            SIOConnectionError,
            SocketIOError,
        ) as err:
            raise WebsocketError(err) from err

    async def disconnect(self) -> None:
        """Disconnect from the socket."""
        await self._sio.disconnect()
        self._watchdog.cancel()

    async def reconnect(self) -> None:
        """Reconnect the websocket connection."""
        await self.disconnect()
        await asyncio.sleep(1)
        await self.connect()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aioambient']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.4,<4.0.0',
 'python-engineio>=3.13.1,<5.0.0',
 'python-socketio>=4.6,<6.0',
 'websockets>=8.1,<10.0']

setup_kwargs = {
    'name': 'aioambient',
    'version': '1.2.6',
    'description': 'A clean, async-friendly library for the Ambient Weather API',
    'long_description': '# 🌤  aioambient: An async library for Ambient Weather Personal Weather Stations\n\n[![CI](https://github.com/bachya/aioambient/workflows/CI/badge.svg)](https://github.com/bachya/aioambient/actions)\n[![PyPi](https://img.shields.io/pypi/v/aioambient.svg)](https://pypi.python.org/pypi/aioambient)\n[![Version](https://img.shields.io/pypi/pyversions/aioambient.svg)](https://pypi.python.org/pypi/aioambient)\n[![License](https://img.shields.io/pypi/l/aioambient.svg)](https://github.com/bachya/aioambient/blob/master/LICENSE)\n[![Code Coverage](https://codecov.io/gh/bachya/aioambient/branch/dev/graph/badge.svg)](https://codecov.io/gh/bachya/aioambient)\n[![Maintainability](https://api.codeclimate.com/v1/badges/81a9f8274abf325b2fa4/maintainability)](https://codeclimate.com/github/bachya/aioambient/maintainability)\n[![Say Thanks](https://img.shields.io/badge/SayThanks-!-1EAEDB.svg)](https://saythanks.io/to/bachya)\n\n`aioambient` is a Python3, asyncio-driven library that interfaces with both the\nREST and Websocket APIs provided by\n[Ambient Weather](https://ambientweather.net).\n\n# Installation\n\n```python\npip install aioambient\n```\n\n# Python Versions\n\n`aioambient` is currently supported on:\n\n* Python 3.7\n* Python 3.8\n* Python 3.9\n\n# API and Application Keys\n\nUtilizing `aioambient` requires both an Application Key and an API Key from\nAmbient Weather. You can generate both from the Profile page in your\n[Ambient Weather Dashboard](https://dashboard.ambientweather.net).\n\n# Usage\n\n## Creating a Client\n\nAn `aioambient` client starts with an\n[aiohttp](https://aiohttp.readthedocs.io/en/stable/) `ClientSession`:\n\n```python\nimport asyncio\n\nfrom aiohttp import ClientSession\n\nfrom aioambient import Client\n\n\nasync def main() -> None:\n    """Create the aiohttp session and run the example."""\n    async with ClientSession() as session:\n        # YOUR CODE HERE\n\n\nasyncio.get_event_loop().run_until_complete(main())\n```\n\nCreate a client, initialize it, then get to it:\n\n```python\nimport asyncio\nfrom datetime import date\n\nfrom aiohttp import ClientSession\n\nfrom aioambient import Client\n\n\nasync def main() -> None:\n    """Create the aiohttp session and run the example."""\n    async with ClientSession() as session:\n        client = Client("<YOUR API KEY>", "<YOUR APPLICATION KEY>", session)\n\n        # Get all devices in an account:\n        await client.api.get_devices()\n\n        # Get all stored readings from a device:\n        await client.api.get_device_details("<DEVICE MAC ADDRESS>")\n\n        # Get all stored readings from a device (starting at a date):\n        await client.api.get_device_details(\n            "<DEVICE MAC ADDRESS>", end_date=date(2019, 1, 16)\n        )\n\n\nasyncio.get_event_loop().run_until_complete(main())\n```\n\n## REST API\n\n```python\nimport asyncio\nfrom datetime import date\n\nfrom aiohttp import ClientSession\n\nfrom aioambient import Client\n\n\nasync def main() -> None:\n    """Create the aiohttp session and run the example."""\n    client = Client("<YOUR API KEY>", "<YOUR APPLICATION KEY>")\n\n    # Get all devices in an account:\n    await client.api.get_devices()\n\n    # Get all stored readings from a device:\n    await client.api.get_device_details("<DEVICE MAC ADDRESS>")\n\n    # Get all stored readings from a device (starting at a datetime):\n    await client.api.get_device_details(\n        "<DEVICE MAC ADDRESS>", end_date=date(2019, 1, 16)\n    )\n\n\nasyncio.run(main())\n```\n\nBy default, the library creates a new connection to Ambient Weather with each coroutine.\nIf you are calling a large number of coroutines (or merely want to squeeze out every\nsecond of runtime savings possible), an\n[`aiohttp`](https://github.com/aio-libs/aiohttp) `ClientSession` can be used for connection\npooling:\n\n```python\nimport asyncio\nfrom datetime import date\n\nfrom aiohttp import ClientSession\n\nfrom aioambient import Client\n\n\nasync def main() -> None:\n    """Create the aiohttp session and run the example."""\n    async with ClientSession() as session:\n        client = Client("<YOUR API KEY>", "<YOUR APPLICATION KEY>", session=session)\n\n        # Get all devices in an account:\n        await client.api.get_devices()\n\n        # Get all stored readings from a device:\n        await client.api.get_device_details("<DEVICE MAC ADDRESS>")\n\n        # Get all stored readings from a device (starting at a datetime):\n        await client.api.get_device_details(\n            "<DEVICE MAC ADDRESS>", end_date=date(2019, 1, 16)\n        )\n\n\nasyncio.run(main())\n```\n\nPlease be aware of Ambient Weather\'s\n[rate limiting policies](https://ambientweather.docs.apiary.io/#introduction/rate-limiting).\n\n## Websocket API\n\n```python\nimport asyncio\n\nfrom aiohttp import ClientSession\n\nfrom aioambient import Client\n\n\nasync def main() -> None:\n    """Create the aiohttp session and run the example."""\n    async with ClientSession() as session:\n        client = Client("<YOUR API KEY>", "<YOUR APPLICATION KEY>", session)\n\n        # Define a method that should be fired when the websocket client\n        # connects:\n        def connect_method():\n            """Print a simple "hello" message."""\n            print("Client has connected to the websocket")\n\n        client.websocket.on_connect(connect_method)\n\n        # Alternatively, define a coroutine handler:\n        async def connect_coroutine():\n            """Waits for 3 seconds, then print a simple "hello" message."""\n            await asyncio.sleep(3)\n            print("Client has connected to the websocket")\n\n        client.websocket.async_on_connect(connect_coroutine)\n\n        # Define a method that should be run upon subscribing to the Ambient\n        # Weather cloud:\n        def subscribed_method(data):\n            """Print the data received upon subscribing."""\n            print(f"Subscription data received: {data}")\n\n        client.websocket.on_subscribed(subscribed_method)\n\n        # Alternatively, define a coroutine handler:\n        async def subscribed_coroutine(data):\n            """Waits for 3 seconds, then print the incoming data."""\n            await asyncio.sleep(3)\n            print(f"Subscription data received: {data}")\n\n        client.websocket.async_on_subscribed(subscribed_coroutine)\n\n        # Define a method that should be run upon receiving data:\n        def data_method(data):\n            """Print the data received."""\n            print(f"Data received: {data}")\n\n        client.websocket.on_data(data_method)\n\n        # Alternatively, define a coroutine handler:\n        async def data_coroutine(data):\n            """Wait for 3 seconds, then print the data received."""\n            await asyncio.sleep(3)\n            print(f"Data received: {data}")\n\n        client.websocket.async_on_data(data_coroutine)\n\n        # Define a method that should be run when the websocket client\n        # disconnects:\n        def disconnect_method(data):\n            """Print a simple "goodbye" message."""\n            print("Client has disconnected from the websocket")\n\n        client.websocket.on_disconnect(disconnect_method)\n\n        # Alternatively, define a coroutine handler:\n        async def disconnect_coroutine(data):\n            """Wait for 3 seconds, then print a simple "goodbye" message."""\n            await asyncio.sleep(3)\n            print("Client has disconnected from the websocket")\n\n        client.websocket.async_on_disconnect(disconnect_coroutine)\n\n        # Connect to the websocket:\n        await client.websocket.connect()\n\n        # At any point, disconnect from the websocket:\n        await client.websocket.disconnect()\n\n\nasyncio.run(main())\n```\n\n# Contributing\n\n1. [Check for open features/bugs](https://github.com/bachya/aioambient/issues)\n  or [initiate a discussion on one](https://github.com/bachya/aioambient/issues/new).\n2. [Fork the repository](https://github.com/bachya/aioambient/fork).\n3. (_optional, but highly recommended_) Create a virtual environment: `python3 -m venv .venv`\n4. (_optional, but highly recommended_) Enter the virtual environment: `source ./venv/bin/activate`\n5. Install the dev environment: `script/setup`\n6. Code your new feature or bug fix.\n7. Write tests that cover your new functionality.\n8. Run tests and ensure 100% code coverage: `script/test`\n9. Update `README.md` with any new documentation.\n10. Add yourself to `AUTHORS.md`.\n11. Submit a pull request!\n',
    'author': 'Aaron Bach',
    'author_email': 'bachya1208@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bachya/aioambient',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
